/*{*************************************************************************
                          XMBAtom.h  -  description
                             -------------------
begin                : Fri Jul 21 2006
copyright            : (C) 2006 by ImazInatIon
email                : revanth_nr@students.iiit.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *************************************************************************}*/

#ifndef XMBATOM_H
#define XMBATOM_H

#include <qcanvas.h>
#include "XMBBond.h"
#include "XMBMolecule.h"
#include <qtextstream.h>
#include <qvaluelist.h>
#include <qpair.h>

class XMBBond;
class XMoleculeBuilderDoc;

/**
 *@author Revanth
 */

class XMBAtom : public QCanvasEllipse  {
	public:

		// constructor
		XMBAtom(QCanvas *canvas, float _atomz, bool lock=0);
		XMBAtom(QCanvas *canvas, QString atomName, float _atomz, bool lock=0);

		// destructor
		~XMBAtom();

		// add inbond - ie this is not the starting point
		void addInBond( XMBBond *bond );

		// add outbond - ie this is the starting atom for the bond
		void addOutBond( XMBBond *bond );

		// delete bond
		bool deleteBond( XMBBond *bond);

		// transfer bonds
		void transferBonds( XMBAtom *atom);

		// combine
		bool combine( XMBAtom *a);

		// set molecule
		void setMolecule( XMBMolecule *m);

		// get molecule
		XMBMolecule *getMolecule(void);

		// move the atom
		void move( double dx, double dy);
		void moveBy( double dx, double dy);

		// rotate the atom
		void rotateAbout( double x, double y, float phi );

		// check if the atom is connected to given other atom and returns the connecting bond
		// if it is connected, otherwise return null
		XMBBond  *isConnected( XMBAtom *atom);

		// if there is any path to given atom
		bool  isAnyhowConnected( XMBAtom *atom);
		bool  isInListConnected( XMBAtom *atom);
		bool  isOutListConnected( XMBAtom *atom);

		// get bond count
		int getBondCount();

		// get inlist
		QPtrList<XMBBond> *getInList();

		// get outlist
		QPtrList<XMBBond> *getOutList();

		// getFirstConnectedAtom();
		XMBAtom *getFirstConnectedAtom();

		// write this atom and its connectivity to stream
		void writeMoleculeToFile( QTextStream *s, QValueList< QPair<XMBAtom*,XMBAtom*> > *l);

		// read molecule from file
		void readMoleculeFromFile(QTextStream *s, XMBMolecule *molecule, XMoleculeBuilderDoc *doc);

		// hover effect
		void startHover(void);
		void stopHover(void);

		// name and symbol
		QString getName(void);
		QString getSymbol(void);
		void setName( QString n);
		void setSymbol( QString s);

		// plane of atom
		float getAtomZ(void);
		void setAtomZ(float z);

		int deleted;
		bool locked;

		QString _atomColor[3];

	private:

		// atom name
		QString name;

		// atom symbol
		QString symbol;

		// molecule to which this atom belongs to
		XMBMolecule *molecule;

		// temp flag
		int flag;

		float atomZ;

		// show symbol on canvas
		QCanvasText *canvasSymbol;

		// inlist and outlist stored here
		QPtrList<XMBBond> inList;
		QPtrList<XMBBond> outList;
};

#endif

