/*{**************************************************************************
  XMBFileIO.cpp  -  description
  -------------------
begin                : Fri Jul 21 2006
copyright            : (C) 2006 by ImazInatIon
email                : revanth_nr@students.iiit.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *************************************************************************}*/


#include <iostream>
#include <cmath>
#include "XMBAtom.h"
#include "XMBBond.h"
#include "XMBMolecule.h"
#include "XMBFileIO.h"
#include "xmoleculebuilderdoc.h"

XMBFileIO::XMBFileIO(){};
XMBFileIO::~XMBFileIO(){};

int XMBFileIO::getIndex( XMBAtom *atom){
	int i;
	for( i=0; i< atomsCount; i++){
		if( atomsArr[i] == atom )
			return i;
	}
	return 0;
}

void XMBFileIO::writeMolecule( QTextStream &s, XMBMolecule *molecule, QCanvas *canvas){

	QCanvasItemList::iterator it;
	QCanvasItemList l = canvas->allItems();
	XMBBond *bond;
	int i,j;

	bondsCount = atomsCount = 0;

	for( it=l.begin(); it != l.end(); it++){
		if( (*it)->rtti() == 6 && ((XMBAtom*)(*it))->getMolecule() == molecule ){
			atomsCount++;
			bondsCount += ((XMBAtom*)(*it))->getBondCount();
		}
	}
	bondsCount /= 2;
	s<<" MOLECULE "<<atomsCount<<" "<<bondsCount<<endl;

	// get list of atoms and moelcules
	atomsArr = new XMBAtom*[atomsCount];
	bondsArr = new XMBBond**[atomsCount];
	for(i=0; i< atomsCount; i++){
		bondsArr[i] = new XMBBond*[atomsCount];
		for( j=0; j<atomsCount; j++)
			bondsArr[i][j] = NULL;
	}
	i=0;
	for( it=l.begin(); it != l.end(); it++){
		if( (*it)->rtti() == 6 && ((XMBAtom*)(*it))->getMolecule() == molecule ){
			atomsArr[i++] = ((XMBAtom*)(*it));
		}
	}
	for( it=l.begin(); it != l.end(); it++){
		if( (*it)->rtti() == 6 && ((XMBAtom*)(*it))->getMolecule() == molecule ){

			QPtrListIterator<XMBBond> bit1( *((XMBAtom*)(*it))->getInList() );
			while( bond = bit1.current() ){
				++bit1;
				bondsArr[ getIndex( bond->getFromAtom() ) ][ getIndex( bond->getToAtom() ) ] = bond;
			}
			QPtrListIterator<XMBBond> bit2( *((XMBAtom*)(*it))->getOutList() );
			while( bond = bit2.current() ){
				++bit2;
				bondsArr[ getIndex( bond->getFromAtom() ) ][ getIndex( bond->getToAtom() ) ] = bond;
			}
		}
	}

	// write the atoms and molecules
	for( i=0; i<atomsCount; i++){
		s<<"  ATOM "<<atomsArr[i]->getSymbol()<<" "<<atomsArr[i]->x()<<" "<<atomsArr[i]->y()<<" "<<atomsArr[i]->getAtomZ()<<" "<<atomsArr[i]->_atomColor[0]<<" "<<atomsArr[i]->_atomColor[1]<<" "<<atomsArr[i]->_atomColor[2]<<endl;
	}
	for( i=0; i<atomsCount; i++){
		for( j=0; j<atomsCount; j++){
			if( bondsArr[i][j] != NULL ){
				s<<"  BOND "<<i<<" "<<j<<" "<<bondsArr[i][j]->bondCount<<" ";
				if(bondsArr[i][j]->bondCount==1)
					s<<bondsArr[i][j]->bond1Type<<endl;
				else if(bondsArr[i][j]->bondCount==2)
					s<<bondsArr[i][j]->bond1Type<<" "<<bondsArr[i][j]->bond2Type<<endl;
				else if(bondsArr[i][j]->bondCount==3)
					s<<bondsArr[i][j]->bond1Type<<" "<<bondsArr[i][j]->bond2Type<<" "<<bondsArr[i][j]->bond3Type<<endl;
			}
		}
	}
	s<<" END_MOLECULE"<<endl;

	for( i=0; i<atomsCount; i++)
		delete[] bondsArr[i];
	delete[] bondsArr;
	delete[] atomsArr;

}

void XMBFileIO::writeWorkspace( QTextStream &s, QCanvas *canvas){

	QCanvasItemList::iterator it;
	QCanvasItemList l = canvas->allItems();
	QValueList<XMBMolecule*> l1;
	XMBMolecule *molecule;

	// header
	s<<"XMBFILE"<<endl;

	for( it=l.begin(); it != l.end(); it++){
		// get molecule to write
		if( (*it)->rtti() == 6 && l1.find( ( (XMBAtom*)(*it))->getMolecule() ) == l1.end() ) {
			molecule = ((XMBAtom*)(*it))->getMolecule();
			l1.append( molecule );
			writeMolecule(s, molecule, canvas);
		}

	}

	s<<"END_XMBFILE"<<endl;
}

bool XMBFileIO::readWorkspace( QTextStream &s, XMoleculeBuilderDoc *doc,bool locked){

	QString str,_r,_g,_b;
	XMBBond *bond;
	XMBMolecule *molecule;
	int bondType, i,j, Bcount;
	float x, y, z;
	float angle;
	int a, b;

	s>>str;
	if( str != "XMBFILE" )
		return false;
	while( 1 ){

		s>>str;

		if( str == "END_XMBFILE" )
			return true;

		if( str == "MOLECULE" ){
			molecule = new XMBMolecule();
			s>>atomsCount>>bondsCount;
			atomsArr = new XMBAtom*[atomsCount];
			for( i=0; i<atomsCount; i++){
				s>>str>>str>>x>>y>>z>>_r>>_g>>_b;
				atomsArr[i] = new XMBAtom(doc->canvas,str,z,locked);
				atomsArr[i]->_atomColor[0]=_r;
				atomsArr[i]->_atomColor[1]=_g;
				atomsArr[i]->_atomColor[2]=_b;
				atomsArr[i]->move( x, y);
				atomsArr[i]->setMolecule( molecule );
				atomsArr[i]->show();
			}
			for( i=0; i<bondsCount; i++)
			{
				s>>str>>a>>b>>Bcount;
				bond = new XMBBond( atomsArr[a] , atomsArr[b], doc);
				for(j=0;j<Bcount;j++)
				{
					s>>bondType;
					if(bondType==2 || bondType==3)
					{
						int x1=atomsArr[a]->x();
						int y1_=atomsArr[a]->y();
						int x2=atomsArr[b]->x();
						int y2=atomsArr[b]->y();
						float length=sqrt((x2-x1)*(x2-x1)+(y2-y1_)*(y2-y1_));
						angle=180/M_PI*atan( fabs( (atomsArr[a]->getAtomZ()-atomsArr[b]->getAtomZ())/length ) );
						bond->addBond( (XMBBondType) bondType, angle);
					}
					else
						bond->addBond( (XMBBondType) bondType);
				}
			}
			s>>str;
			delete[] atomsArr;
		}
		else
			return false;

	}
}


