/***************************************************************************
                          xmoleculebuilderview.cpp  -  description
                             -------------------
    begin                : Fri Jul 21 18:57:53 IST 2006
    copyright            : (C) |2006| by Revanth
    email                : revanth_nr@students.iiit.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qprinter.h>
#include <qpainter.h>

// application specific includes
#include "xmoleculebuilderview.h"
#include "xmoleculebuilderdoc.h"
#include <qmessagebox.h>
#include <qapplication.h>
#include <iostream>
#include "XMBAtomProperties.h"
#include "XMBBondProperties.h"

#include <math.h>

using namespace std;

QCanvasLine tempBond(NULL);
double fixedAngleValue, fixedLengthValue, newx, newy;
double _x;

XMoleculeBuilderView::XMoleculeBuilderView(/*XMoleculeBuilderApp*/QWidget* parent, XMoleculeBuilderDoc *pdoc) : QCanvasView(parent)
{
	/** connect doc with the view*/
//	app=parent;
	doc=pdoc;
	connect(doc, SIGNAL(documentChanged()), this, SLOT(slotDocumentChanged()));
	setVScrollBarMode(AlwaysOn);
	setHScrollBarMode(AlwaysOn);
	setBackgroundColor(QColor(128,128,0));      // what is the use ???
	setCursor( Qt::crossCursor );
	setCanvas(doc->canvas);
	tempBond.setCanvas( doc->canvas);
	tempBond.setZ(0);
	fixedAngleValue = M_PI/18.0;
	fixedLengthValue = 30.0;
	fixedAngleFlag = 0;
	fixedLengthFlag = 0;
	flag=0;
	hover = NULL;

	viewport()->setMouseTracking(true);
}

XMoleculeBuilderView::~XMoleculeBuilderView(){
}

void XMoleculeBuilderView::slotDocumentChanged()
{
	//TODO update the view

}


#include <math.h>

/*
 * find angle of inclination of point P1(x1, y1) 
 * with respect to point P0(x0, y0)
 */
float findtheta( float x0, float y0, float x1, float y1){

	float theta;

	if( x0 == x1 )
		theta = M_PI/2.0;
	else
		theta = atan( (y1-y0)/(x1-x0) );
	if( x1 > x0  && y1 > y0 )
		theta = 2.0 * M_PI - theta;
	if( x1 > x0 && y1 < y0 )
		theta *= -1.0;
	if( x1 < x0 )
		theta = M_PI - theta;

	return theta;
}

/*
 * finds the Z value of the cursor position
 * with respect to the bond being drawn
 */
float XMoleculeBuilderView::getZplane( void ) {
	
	if ( canvasState == XMBVS_drawStereoUp)
		return lastAtom->getAtomZ() + sqrt((lastAtom->x()-newx)*(lastAtom->x()-newx) + (lastAtom->y()-newy)*(lastAtom->y()-newy)) * tan(M_PI/3.0);
	else if ( canvasState == XMBVS_drawStereoDown )
		return lastAtom->getAtomZ() - sqrt((lastAtom->x()-newx)*(lastAtom->x()-newx) + (lastAtom->y()-newy)*(lastAtom->y()-newy)) * tan(M_PI/3.0);
	else
		return lastAtom->getAtomZ();
}

/*
 * draw zigzag pattern
 */
void XMoleculeBuilderView::drawZigZag(QMouseEvent *e){
	QPoint p = inverseWorldMatrix().map(e->pos());
	XMBBond *bond;

	/* continuity using ctrl presses */
	if( !(e->state() &  Qt::ControlButton) ){
		isLastAtom = 0;
		tempBond.hide();
	}

	float ix = lastAtom->x(), iy = lastAtom->y();
	float anglebetween = 30.0; // TODO:replace (180.0 - fixedAngleValue)/2;;
	float angleinclination = 0.0;
	float numbonds;
	XMBAtom *a, *b;
	angleinclination = findtheta( ix, iy, newx, newy ) ;
	angleinclination =  angleinclination * 180.0 / 3.142;
	numbonds = sqrt( (newx-ix)*(newx-ix)+(newy-iy)*(newy-iy) )/(2.0*fixedLengthValue*cos(anglebetween*3.142/180.0)) - 0.5;

	for(int bondcount = 0; bondcount < numbonds; bondcount++){
		// TODO:delete cout<<"here drawing next atoms.... "<< numbonds<<" "<<ix <<" "<<iy<<" "<<angleinclination<<endl;
		a = new XMBAtom(doc->canvas, lastAtom->getAtomZ());
		a->moveBy( ix + fixedLengthValue * cos( (anglebetween + angleinclination)*3.142/180.0) , iy - fixedLengthValue * sin( (anglebetween + angleinclination)*3.142/180.0 ));
		a->show();
		if( bondcount > 0){
			bond = new XMBBond( b, a, doc );
			bond->addBond( canvasState == XMBVS_drawDashedBond ? XMBBT_DashedLine : XMBBT_SingleBond );
			b->setMolecule( a->getMolecule() );
		}else{
			bond = new XMBBond( lastAtom, a, doc );
			bond->addBond( canvasState == XMBVS_drawDashedBond ? XMBBT_DashedLine : XMBBT_SingleBond );
			a->setMolecule( lastAtom->getMolecule() );
		}
		lastAtom = a;
		if( numbonds - bondcount > 0.5 ){
			b = new XMBAtom(doc->canvas, lastAtom->getAtomZ());
			b->moveBy( ix = ix + 2 * fixedLengthValue * cos(anglebetween*3.142/180.0) * cos(angleinclination*3.142/180.0) , iy = iy - 2 * fixedLengthValue * cos(anglebetween*3.142/180.0) * sin(angleinclination*3.142/180.0) );
			b->show();
			bond = new XMBBond( a, b, doc );
			bond->addBond( canvasState == XMBVS_drawDashedBond ? XMBBT_DashedLine : XMBBT_SingleBond );
			b->setMolecule( a->getMolecule() );
			lastAtom = b;
		}
	}
	if( numbonds - fabs(numbonds) > 0.5 ){
		a = new XMBAtom(doc->canvas, lastAtom->getAtomZ());
		a->moveBy( ix + fixedLengthValue * cos( (anglebetween + angleinclination)*3.142/180.0) , iy - fixedLengthValue * sin( (anglebetween + angleinclination)*3.142/180.0 ));
		a->show();
		bond = new XMBBond( b, a, doc );
		bond->addBond( canvasState == XMBVS_drawDashedBond ? XMBBT_DashedLine : XMBBT_SingleBond );
		a->setMolecule( b->getMolecule() );
		lastAtom = a;
	}
}

/* 
 * add bond dependig on canvas state 
 */
void XMoleculeBuilderView::addBond( XMBBond *bond, XMBViewState vs ){

	switch( vs ){
		case XMBVS_drawSimpleBond:
			bond->addBond( XMBBT_SingleBond );
			break;
		case XMBVS_drawDashedBond:
			bond->addBond( XMBBT_DashedLine );
			break;
		case XMBVS_drawStereoUp:
			bond->addBond( XMBBT_StereoUp );
			break;
		case XMBVS_drawStereoDown:
			bond->addBond( XMBBT_StereoDown );
			break;
	}
}


/*
 * draw bond
 */
void XMoleculeBuilderView::drawBond(QMouseEvent *e){

	QPoint p = inverseWorldMatrix().map(e->pos());
	XMBBond *bond;
	QCanvasItemList::Iterator it;
	QCanvasItemList l;

	/* continuity */
	if( !(e->state() &  Qt::ControlButton) ){
		isLastAtom = 0;
		tempBond.hide();
	}

	l=canvas()->collisions(p);
	for (it=l.begin(); it!=l.end(); ++it){
		/* end point is atom */
		if( (*it)->rtti() == 6){  //MoleculeBuilderAtom::atomRTTI
			if( lastAtom == *it )
			{
				flag=1;
				continue;
			}
			if( fabsf( ((XMBAtom*)(*it))->getAtomZ() - getZplane() ) > 0 )
				continue;
			/* if there is already a bond then use add bond otherwise create a new bond	*/
			bond = lastAtom->isConnected( (XMBAtom *)(*it) );
			if( bond == NULL ) {
				bond = new XMBBond( lastAtom, (XMBAtom *)(*it), doc );
				addBond( bond, canvasState );
				((XMBAtom *)(*it))->setMolecule( lastAtom->getMolecule() );
			}
			else {
				addBond( bond, canvasState );
			}
			if( e->state() &  Qt::ControlButton )
				lastAtom = (XMBAtom *)(*it);
			break;
		}
	}

	/* check for collision at proposed point */
	if( it == l.end() ){
		l=canvas()->collisions( QPoint( newx, newy ) );
		for (it=l.begin(); it!=l.end(); ++it){
			/* if end point is atom */
			if( (*it)->rtti() == 6){			//MoleculeBuilderAtom::atomRTTI
				if( lastAtom == *it )
				{
					flag=1;
					continue;
				}
				if( fabsf( ((XMBAtom*)(*it))->getAtomZ() - getZplane() ) > 0 )
					continue;
				/* if there is already a bond then use add bond otherwise create a new bond */
				bond = lastAtom->isConnected( (XMBAtom *)(*it) );
				if( bond == NULL ){
					bond = new XMBBond( lastAtom, (XMBAtom *)(*it), doc );
					addBond( bond, canvasState );
					((XMBAtom *)(*it))->setMolecule( lastAtom->getMolecule() );
				}
				else{
					addBond( bond, canvasState );
				}
				if( e->state() &  Qt::ControlButton )
					lastAtom = (XMBAtom *)(*it);
				break;
			}
		}
	}

	/* if endpoint is not an atom then create a new atom there */
	if( it == l.end() ){
		if( !flag ) {
			XMBAtom *newAtom = new XMBAtom(doc->canvas,getZplane());
			newAtom->setMolecule( lastAtom->getMolecule() );
			newAtom->move(newx, newy);
			newAtom->show();
			bond = new XMBBond( lastAtom, newAtom, doc );
			addBond( bond, canvasState );
			if( e->state() &  Qt::ControlButton )
				lastAtom = newAtom;
		}
		flag=0;
	}
}

/*
 * actions to be taken on mouse press event on the canvas
 * - RightButton click
 *   - stop moving
 * - LeftButton click
 * - If delete mode then delete the atom or bond or molecule
 *  - drawSimpleBond or drawDashedBond
 *  - select the moving object
 */
void XMoleculeBuilderView::contentsMousePressEvent( QMouseEvent *e){
	QPoint p = inverseWorldMatrix().map(e->pos());
	QCanvasItemList::Iterator it;
	QCanvasItemList l;
	XMBBond *bond;

	/* on right click show some options */
	/* moving should be stopped on right click */
	if( e->button() ==  Qt::RightButton ){
		moving = 0;
		l=canvas()->collisions(p);
		// popup if there is some element atom or bond
		// now do some processing accoringly
		for(it  = l.begin(); it!= l.end(); it++){
//			printf("%d\n",(*it)->rtti());
			if( (*it)->rtti() == 6 )
			{
				XMBAtomProperties *pAtomProperties = new XMBAtomProperties;
				pAtomProperties->setAtom( (XMBAtom *)(*it));
				pAtomProperties->exec();
				delete pAtomProperties;
			}
			else if( (*it)->rtti()==4 )
			{
				XMBBondProperties *pBondProperties = new XMBBondProperties;
				pBondProperties->setBond( (XMBStereoBond *)(*it));
				pBondProperties->exec();
				delete pBondProperties;
			}
		}
	}

	/* on left click */
	else if( e->button() == Qt::LeftButton ){

		/* delete action */
		if( canvasState == XMBVS_deleteItem ){
			stopHover();
			l=canvas()->collisions(p);
			it = l.begin();
			if(it!=l.end() && (*it)->rtti() != 3){
				if( (*it)->rtti() == 4)
					( (XMBStereoBond *)(*it) )->deleteBond();
				else
					delete (*it);
			}
		}

		if( isLastAtom == 1 ){
			/* draw zigzag pattern */
			if(canvasState == XMBVS_zigzag )
			{
				drawZigZag(e);
				emit zigzagAtomValue(0);
//				app->statusBar()->clear();
			}
			/* draw bond */
			else if(canvasState == XMBVS_drawSimpleBond || canvasState == XMBVS_drawDashedBond || canvasState == XMBVS_drawStereoUp || canvasState == XMBVS_drawStereoDown)
				drawBond(e);
		}
		else{
			/* if this is my first click then store the current atom ay lastAtom */
			if( canvasState == XMBVS_zigzag || canvasState == XMBVS_drawSimpleBond || canvasState == XMBVS_drawDashedBond || canvasState == XMBVS_drawStereoUp || canvasState == XMBVS_drawStereoDown){
				isLastAtom = 1;
				newx = p.x();
				newy = p.y();
				tempBond.setPoints( 0, 0, 0, 0);
				tempBond.show();
				l=canvas()->collisions(p);
				for (it=l.begin(); it!=l.end(); ++it){
					if( (*it)->rtti() == 6){				//MoleculeBuilderAtom::atomRTTI
						lastAtom = (XMBAtom*)*it;
						break;
					}
				}
				/* if no collision then create a new atom there */
				if( it == l.end() ){
					lastAtom = new XMBAtom(doc->canvas,0);
					lastAtom->move(e->x(),e->y());
					lastAtom->show();
				}
			}
		}

		/* view mode */
		if( canvasState == XMBVS_view ){
			QPoint p = inverseWorldMatrix().map(e->pos());
			QCanvasItemList l=canvas()->collisions(p);
			for (it=l.begin(); it!=l.end(); ++it) {
				moving = *it;
				moving_start = p;
				return;
			}
			moving = 0;
		}

		/* rotate mode */
		if( canvasState == XMBVS_rotate ){
			QPoint p = inverseWorldMatrix().map(e->pos());
			QCanvasItemList l=canvas()->collisions(p);
			for (it=l.begin(); it!=l.end(); ++it) {
				if( (*it)->rtti() != 6 )
					continue;
				moving = *it;
				_x=moving->x();
				return;
			}
			moving = 0;
		}
	}
}

/* actions to invoke on Mouse Release event on canvas
 * - drop the moving object
 *   -if atom is dropped on another atom then join them
 */
void XMoleculeBuilderView::contentsMouseReleaseEvent( QMouseEvent *e){

	QPoint p = inverseWorldMatrix().map(e->pos());
	QCanvasItemList::Iterator it;
	QCanvasItemList l;

	// if moving atom droped on another atom, then delete the moving atom and add its bond to the other atom
	// cannot drop an atom to its only connection because that is equal to deleting that atom
	// REASON - cannot solve deleting atom with bond to itself problem :(
	if( canvasState == XMBVS_view && moving && moving->rtti() == 6){
		stopHover();
		l=canvas()->collisions(p);
		for (it=l.begin(); it!=l.end(); ++it){
			if( (*it)->rtti() == 6 && (*it) != moving ){
				// change it depending on what is preserved
				// the moving atom or the destination
				//((XMBAtom *)moving)->combine( (XMBAtom*)(*it) );
				if( ((XMBAtom *)(*it))->combine( (XMBAtom*)moving ) )
					delete moving;		// if i want to delete both atoms in o---o combining situation
				break;
			}
		}
	}
	/* drop the moving object */
	moving=0;
}

/*
 * start hover
 * must hide the tempbond before selecting the item to hover
 */
void XMoleculeBuilderView::startHover(QPoint p){
	QCanvasItemList::Iterator it;
	QCanvasItemList l;
	if( tempBond.isVisible() ){
		tempBond.hide();
		l=canvas()->collisions(p);
		tempBond.show();
	}else
		l=canvas()->collisions(p);

	it=l.begin();

	if( it!= l.end() && *it != NULL){
		hover = *it;
		if( hover->rtti() == 6)
			((XMBAtom *)(hover))->startHover();
		else if( hover->rtti() == 7)
			((XMBBondLine*)(hover))->startHover();
	}
}

/*
 * stop hover effect on last hover object
 */
void XMoleculeBuilderView::stopHover(void){
	if( hover != NULL ){
		if( hover->rtti() == 6 )
			((XMBAtom *)(hover))->stopHover();
		else if( hover->rtti() == 7)
			((XMBBondLine*)(hover))->stopHover();
	}
	hover = NULL;
}

/* actions to invoke on mouse move event on canvas
 * - if drawing then draw next atom
 * - if moving then move current object
 * - show hover effect on the object under mouse
 */
void XMoleculeBuilderView::contentsMouseMoveEvent(QMouseEvent* e){

	QCanvasItemList::Iterator it;
	QCanvasItemList l;
	QPoint p = inverseWorldMatrix().map(e->pos());

	/* mouse hover */
	l=canvas()->collisions(p);
	it=l.begin();
	if( it == l.end() || *it != hover){
		stopHover();
		startHover(p);
	}

	/* draw state */
	if(canvasState == XMBVS_drawSimpleBond || canvasState == XMBVS_drawDashedBond || canvasState == XMBVS_zigzag || canvasState == XMBVS_drawStereoUp || canvasState == XMBVS_drawStereoDown){


		/* if the from atom is selected */
		/* then predict the position of end atom according to fixedAngle and fixedLength rules */
		if( isLastAtom ){
			float newtheta;
			newx = p.x();
			newy = p.y();
			newtheta = findtheta( lastAtom->x(), lastAtom->y(), p.x(), p.y() );

			// set for fixed angle
			//	    / (x2,y2)
			//	   /
			//(x0,y0) /) <- thetha
			//	  \
			//	   \ (x1,y1)
			if( fixedAngleFlag ){
				float oldtheta, mytheta;
				float newdx, newdy, newr;
				int tempFactor;

				XMBAtom *oa = lastAtom->getFirstConnectedAtom();
				mytheta = findtheta( lastAtom->x(), lastAtom->y(), p.x(), p.y() );
				if( oa == NULL )
					oldtheta = 0;
				else
					oldtheta = findtheta( lastAtom->x(), lastAtom->y(), oa->x(), oa->y() );

				tempFactor = round( fabs( oldtheta - mytheta ) / fixedAngleValue );
				if( oldtheta > mytheta )
					tempFactor *= -1.0;

				newr = pow( (p.x() - lastAtom->x()) * (p.x() -lastAtom->x() ) + ( p.y() - lastAtom->y() ) * ( p.y() -lastAtom->y() ) , 0.5);
				newtheta = oldtheta + tempFactor * fixedAngleValue;
				newx = lastAtom->x() + newr * cos( newtheta );
				newy = lastAtom->y() - newr * sin( newtheta );
			}
			if( fixedLengthFlag && canvasState != XMBVS_zigzag){
				newx = lastAtom->x() + fixedLengthValue * cos( newtheta );
				newy = lastAtom->y() - fixedLengthValue * sin( newtheta );
			}
			tempBond.setPoints( lastAtom->x(), lastAtom->y(), newx, newy);

			if( canvasState == XMBVS_zigzag ) //p || canvasState == XMBVS_drawStereoDown )
			{
				float ix = lastAtom->x(), iy = lastAtom->y();
				float anglebetween = 30.0; // TODO:replace (180.0 - fixedAngleValue)/2;;
				emit zigzagAtomValue(2*(sqrt( (newx-ix)*(newx-ix)+(newy-iy)*(newy-iy) )/(2.0*fixedLengthValue*cos(anglebetween*3.142/180.0)) + 0.5));
			}
			else if( canvasState == XMBVS_drawSimpleBond || canvasState == XMBVS_drawDashedBond || canvasState == XMBVS_drawStereoUp || canvasState == XMBVS_drawStereoDown )
			{
				float ix = lastAtom->x(), iy = lastAtom->y();
				float bondLength = sqrt( (newx-ix)*(newx-ix) + (newy-iy)*(newy-iy) );
				if( canvasState == XMBVS_drawStereoUp || canvasState == XMBVS_drawStereoDown )
					bondLength/=cos(M_PI/3.0);
				float anglebetween;
				if(newx-ix > 0)
					if(newy-iy < 0)
						anglebetween = atan((iy-newy)/(newx-ix));
					else
						anglebetween = 2*M_PI + atan((iy-newy)/(newx-ix));
				else if(newx-ix < 0)
					if(newy-iy < 0)
						anglebetween = M_PI + atan((iy-newy)/(newx-ix));
					else
						anglebetween = M_PI + atan((iy-newy)/(newx-ix));
				else if(newy-iy < 0)
						anglebetween = M_PI/2.0;
					else
						anglebetween = 3*M_PI/2.0;
					
				char bondDetail[50];
				sprintf(bondDetail,"Bond Length: %.3f  Bond Angle: %.2f",bondLength,fmod(anglebetween*180/M_PI,360));
				emit displayBondDetail(bondDetail);
			}
		}
	}

	/* view mode */
	if( canvasState == XMBVS_view ){
		if ( moving ) {
			/* mouse move with control key pressed moves the full molecule */
			if( moving->rtti() == 6 && (e->state() & Qt::ControlButton || ((XMBAtom*)moving)->locked)){
				XMBMolecule* molecule=((XMBAtom*)moving)->getMolecule();
				l=canvas()->allItems();
				for (it=l.begin(); it!=l.end(); ++it){
					if( (*it)->rtti() == 6 && ((XMBAtom*)(*it))->getMolecule() == molecule )
						(*it)->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
				}
			}
			else if( moving->rtti() == 4 && (e->state() & Qt::ControlButton || ((XMBStereoBond*)moving)->getBond()->getFromAtom()->locked) ){
				XMBMolecule* molecule=((XMBStereoBond*)moving)->getBond()->getMolecule();
				l=canvas()->allItems();
				for (it=l.begin(); it!=l.end(); ++it){
					if( (*it)->rtti() == 6 && ((XMBAtom*)(*it))->getMolecule() == molecule )
						(*it)->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
				}
			}
			else if( moving->rtti() == 7 && (e->state() & Qt::ControlButton || ((XMBBondLine*)moving)->getBond()->getFromAtom()->locked) ){
				XMBMolecule* molecule=((XMBBondLine*)moving)->getBond()->getMolecule();
				l=canvas()->allItems();
				for (it=l.begin(); it!=l.end(); ++it){
					if( (*it)->rtti() == 6 && ((XMBAtom*)(*it))->getMolecule() == molecule )
						(*it)->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
				}
			}
			/* otherwise move the current item only */
			else
				moving->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
			
			moving_start = p;
		}
	}
	if( canvasState == XMBVS_rotate ){
		if( moving && moving->rtti() == 6){
			l=canvas()->allItems();
			for (it=l.begin(); it!=l.end(); ++it){
				if( *it != moving && (*it)->rtti() == 6 && ((XMBAtom*)(*it))->getMolecule() == ((XMBAtom*)moving)->getMolecule() ){
					((XMBAtom*)(*it))->rotateAbout( moving->x(), moving->y(), ( p.x() - _x )/10.0);// - (((p.x() - moving->x())/100.0)/360) * 360 );
					//((XMBAtom*)(*it))->rotateAbout( moving->x(), moving->y(), sqrt( float (( p.x() - moving->x() )* (p.x() - moving->x() )+ (p.y() - moving->y()) * ( p.y() - moving->y() ) ) )/50.0  );
					//break;
				}
			}
			_x=p.x();
		}
	}
}
